using System;
using System.Text;
using System.Data;
using System.Collections;
using ERModel;
using NHibernate.Dialect;
using Ormtypemapper;

namespace ERToDB
{
	/// <summary>
	/// Summary description for TableDDLGenerator.
	/// </summary>
	public class TableDDLGenerator
	{
		public TableDDLGenerator()
		{
		}

		public string generateCreateDDL(EREntity entity, Dialect dialect)
		{
			StringBuilder sb = new StringBuilder( "create table " );
			int quoteSqlIdentifier = entity.GetModel().getQuoteSQLIdentifier();
			sb.Append(DDLGeneratorUtil.GetQualifiedName(entity.Schema, entity.GetName(), dialect, quoteSqlIdentifier));
			sb.Append(" (");

			System.Collections.ArrayList uniqueKeys = new System.Collections.ArrayList();
			System.Collections.IEnumerator elementsEnum = entity.GetElementsEnumerator();
			int count = 0;
			while(elementsEnum.MoveNext())
			{
				ERElement element = (ERElement)elementsEnum.Current;
				if(element is ERColumn)
				{
					ERColumn column = element as ERColumn;
					if(count != 0)
					{
						sb.Append(", ");
					}
					sb.Append(DDLGeneratorUtil.QuoteName(column.GetName(), dialect, quoteSqlIdentifier));
					sb.Append(' ');
					int databaseType = DialectDBTypeMapper.Instance().GetDBType(dialect.GetType().Name);

					if(column.PrimaryKey && IsIdentityColumn(column, dialect))
					{
						if (dialect.HasDataTypeInIdentityColumn) 
						{
							sb.Append(GetColumnTypeString(column, databaseType, dialect));
						}
						DBType dbType = DBTypeUtil.getDBTypeByID(column.Type);
						sb.Append(' ').Append(dialect.IdentityColumnString);
					}
					else
					{
						sb.Append(GetColumnTypeString(column, databaseType, dialect));
						
						//default value
						String lDefaultValue = quoteDefaultValue(column, dialect);
						if (lDefaultValue != null && lDefaultValue.Length > 0) 
						{
							sb.Append( " default " ).Append(lDefaultValue);
						}

						if(column.Nullable && !column.PrimaryKey)
						{
							sb.Append(dialect.NullColumnString);
						}
						else
						{
							sb.Append(" not null");
						}
					}

					if(column.Unique)
					{
						if(dialect.SupportsUnique)
						{
							sb.Append(" unique");
						}
						else
						{
							uniqueKeys.Add(column.GetName());
						}
					}

					if(column.HasCheckConstraint() && dialect.SupportsColumnCheck())
					{
						sb.Append(" check(").Append(column.CheckConstraint).Append(")");
					}

					string columnComment = column.Comment;
					if (columnComment != null && columnComment.Length > 0) 
					{
						sb.Append(dialect.GetColumnComment(columnComment));
					}
				}
				count++;
			}//while

			bool foundPK = false;
			foreach(ERElement element in entity.Elements)
			{
				if(element is ERColumn)
				{
					ERColumn column = element as ERColumn;
					if(column.PrimaryKey)
					{
						if(!foundPK)
						{
							string lPKConstraintName = entity.PKConstraintName;
							sb.Append(", ");
							if(lPKConstraintName != null)
							{
								sb.Append(" constraint ");
								sb.Append(DDLGeneratorUtil.QuoteName(lPKConstraintName, dialect, quoteSqlIdentifier));
							}
							sb.Append(" primary key (");
							foundPK = true;
						}
						else
						{
							sb.Append(", ");
						}
						sb.Append(DDLGeneratorUtil.QuoteName(column.GetName(), dialect, quoteSqlIdentifier));
					}
				}
			}//foreach
			if(foundPK)
			{
				sb.Append(")");
			}
	
			System.Collections.IEnumerator uniqueKeyEnum =  entity.uniqueKeysEnumerator();
			while(uniqueKeyEnum.MoveNext())
			{
				sb.Append(",");
				string uniqueKey = (string)uniqueKeyEnum.Current;
				sb.Append(" constraint ");
				sb.Append(DDLGeneratorUtil.QuoteName(uniqueKey, dialect, quoteSqlIdentifier));
				sb.Append(" unique (");
				System.Collections.IEnumerator uniqueColumnEnum = entity.UniqueKeyColumnsEnumerator(uniqueKey);
				int countx = 0;
				while(uniqueColumnEnum.MoveNext())
				{
					if(countx > 0)
					{
						sb.Append(", ");
					}
					sb.Append(DDLGeneratorUtil.QuoteName((string)uniqueColumnEnum.Current, dialect, quoteSqlIdentifier));
					countx++;	

				}
				sb.Append(")");
			}//UniqueKeyEnum
			
			System.Collections.IEnumerator singleUniqueKeyEnum = uniqueKeys.GetEnumerator();
			int single  = 0;
			while(singleUniqueKeyEnum.MoveNext())
			{	
				if(single > 0)
				{
					sb.Append(", " );
				}
				sb.Append(", unique (");
				sb.Append(DDLGeneratorUtil.QuoteName((string)singleUniqueKeyEnum.Current, dialect, quoteSqlIdentifier));
				sb.Append(")");
				single++;
			}

			if(dialect.SupportsColumnCheck())
			{
				System.Collections.IEnumerator checkKeysEnum = entity.CheckConstraintKeysEnumerator();
				while(checkKeysEnum.MoveNext())
				{
					sb.Append(",");
					sb.Append(" constraint ");
					string lName = checkKeysEnum.Current as string;
					sb.Append(DDLGeneratorUtil.QuoteName(lName, dialect, quoteSqlIdentifier));
					sb.Append(" check(").Append(entity.GetCheckConstraints(lName)).Append(")");
				}
			}

			sb.Append(")");

			if (entity.Comment!=null && entity.Comment.Length > 0) 
				sb.Append( dialect.GetTableComment(entity.Comment) );

			sb.Append(dialect.GetTableTypeString());

			if (entity.ExtraDDLClause != null && entity.ExtraDDLClause.Length > 0)
				sb.Append(" ").Append(entity.ExtraDDLClause);

			return sb.ToString();
		}

		public String GenerateIndex(EREntity entity, Dialect dialect, string indexName)
		{
			StringBuilder sb = new StringBuilder( "CREATE INDEX ");
			int quoteSqlIdentifier = entity.GetModel().getQuoteSQLIdentifier();
			sb.Append(DDLGeneratorUtil.QuoteName(indexName, dialect, quoteSqlIdentifier));
			sb.Append(" ON ");
			sb.Append(DDLGeneratorUtil.QuoteName(entity.GetName(), dialect, quoteSqlIdentifier));
			sb.Append("(");
			System.Collections.IEnumerator indexColumnEnum = entity.indexKeyColumnsEnumerator(indexName);
			int count = 0;
			while(indexColumnEnum.MoveNext())
			{
				if(count > 0)
				{
					sb.Append(",");
				}

				sb.Append(DDLGeneratorUtil.QuoteName((string)indexColumnEnum.Current, dialect, quoteSqlIdentifier));

				count++;
			}
			sb.Append(")");

			return sb.ToString();
		}

		private bool IsIdentityColumn(ERColumn column, Dialect dialect) 
		{
			String lIdGenerator = DDLGeneratorUtil.GetIDGenerator(column.IdGenerator, dialect,
				DBTypeUtil.isIntType(column.Type));
			return (lIdGenerator != null && lIdGenerator.Equals(ERModelConstants.IDG_IDENTITY));
		}

		private String GetColumnTypeString(ERColumn aColumn, int aDatabaseType, Dialect aDialect) 
		{
			string lUserType = aColumn.GetUserType(aDatabaseType);
			if (lUserType != null) 
			{
				return lUserType;
			}
		

			DBType lDBType = DBTypeUtil.getDBTypeByID(aColumn.Type);
			string lType = lDBType.getDBType(aDatabaseType);

			if (aDatabaseType == DBType.DBTYPE_SYBASE) 
			{
				if (aColumn.PrimaryKey && !lType.Equals("numeric") && IsIdentityColumn(aColumn, aDialect)) 
				{
					// Sybase identity must be numeric(x, 0)
					return "numeric(19,0)";
				} 
				else if (aColumn.ForeignKey) 
				{
					ERColumn lColumn = aColumn.GetModel().GetColumnByFullName(aColumn.ReferencedColumn);
					if (lColumn != null) 
					{
						DBType lDBType2 = DBTypeUtil.getDBTypeByID(aColumn.Type);
						string lType2 = lDBType2.getDBType(aDatabaseType);
						if (!lType2.Equals("numeric") && IsIdentityColumn(lColumn, aDialect)) 
						{
							// if foreign key reference primary key is identity and not numeric
							return "numeric(19,0)";
						}
					}
				}
			} 
			else if (aDatabaseType == DBType.DBTYPE_INFORMIX) 
			{
				if (lType.Equals("informix_float")) 
				{
					lType = "float";
				}
			}
    	
			int lLength = lDBType.getDBLength(aDatabaseType);
			if (lLength != 0) 
			{
				int lColLength = aColumn.Length;
				if (lColLength == 0) 
				{
					// Maybe changed from MSSQL int(0) to number, get default
					lColLength = lDBType.DefaultLength;
				}
				string lLengthSpecifier = "(";
				lLengthSpecifier += lColLength;
				if (lDBType.Decimal) 
				{
					lLengthSpecifier += "," + aColumn.Scale;
				}
				lLengthSpecifier += ")";
				if (lType.IndexOf("($l)") > 0) 
				{
					lType = lType.Replace("\\(\\$l\\)", lLengthSpecifier);
				} 
				else 
				{
					lType += lLengthSpecifier;
				}
			}
			return lType;
		}

		public string GenerateDropSQL(EREntity entity, Dialect dialect) 
		{            
			return dialect.GetDropTableString(DDLGeneratorUtil.GetQualifiedName(entity.Schema, entity.GetName(), dialect, entity.GetModel().getQuoteSQLIdentifier()));
		}

		public String[] generateAlterDDL(EREntity aEntity, DataTable dataTable, Dialect aDialect) 
		{
			int quoteSqlIdentifier = aEntity.GetModel().getQuoteSQLIdentifier();
			ArrayList lSQL = new ArrayList();
			StringBuilder sb = new StringBuilder("alter table ").Append(
				DDLGeneratorUtil.GetQualifiedName(aEntity.Schema, aEntity.GetName(), aDialect, quoteSqlIdentifier)).Append(' ').Append(aDialect.AddColumnString);
			foreach(ERElement lElement in aEntity.Elements)
			{
				if (lElement is ERColumn) 
				{
					ERColumn lColumn = (ERColumn) lElement;
					bool exist = false;
					foreach(DataRow row in dataTable.Rows)
					{
						//						DataColumn dataColumn = dataTable.Rows["ColumnName"];
						if(row["ColumnName"].ToString().Equals(lColumn.GetName()))
						{
							exist = true;
							break;
						}
					}

					if (exist == false) 
					{
						int lDatabaseType = DialectDBTypeMapper.Instance().GetDBType(aDialect.GetType().Name);
						// the column doesnt exist at all.
						StringBuilder lAlter = new StringBuilder(sb.ToString()).Append(' ').Append(
							DDLGeneratorUtil.QuoteName(lColumn.GetName(), aDialect, quoteSqlIdentifier)).Append(' ').Append(
							GetColumnTypeString(lColumn, lDatabaseType, aDialect));

						//default value
						String lDefaultValue = quoteDefaultValue(lColumn, aDialect);
						if (lDefaultValue != null && lDefaultValue.Length > 0) 
						{
							lAlter.Append( " default " ).Append(lDefaultValue);
						}
						if(lColumn.Nullable)
						{
							lAlter.Append(aDialect.NullColumnString);
						}
						else
						{
							lAlter.Append(" not null");
						}

						if (lColumn.Unique && aDialect.SupportsUnique) 
						{
							lAlter.Append(" unique");
						}
						if (lColumn.HasCheckConstraint()) 
						{
							lAlter.Append(" check(").Append(lColumn.CheckConstraint).Append(")");
						}
						string columnComment = lColumn.Comment;
						if (columnComment != null && columnComment.Length > 0) 
						{
							lAlter.Append(aDialect.GetColumnComment(columnComment));
						}
						lSQL.Add(lAlter.ToString());
					}
				}
			}
			return lSQL.ToArray(typeof(string)) as string[];
		}

		private string quoteDefaultValue(ERColumn aColumn, Dialect aDialect) 
		{
			string lDefaultValue = aColumn.DefaultValue;
			if (lDefaultValue != null && lDefaultValue.Length > 0) 
			{
				DBType lDBType = DBTypeUtil.getDBTypeByID(aColumn.Type);
				int lTypeGroup = lDBType.TypeGroup;
				if (lTypeGroup != DBType.GROUP_INTEGER && lTypeGroup != DBType.GROUP_LONG
					&& lTypeGroup != DBType.GROUP_FLOAT
					&& lTypeGroup != DBType.GROUP_DOUBLE
					&& lTypeGroup != DBType.GROUP_DECIMAL
					&& lTypeGroup != DBType.GROUP_SMALLINT
					&& lTypeGroup != DBType.GROUP_TINYINT)
				{
					if (lDefaultValue.StartsWith("'") && lDefaultValue.EndsWith("'")) 
					{
						return lDefaultValue; // already quoted 
					}
					string lFunctionName = lDefaultValue.ToLower();
					if (lFunctionName.IndexOf('(') > 0) 
					{
						lFunctionName = lFunctionName.Substring(0, lFunctionName.IndexOf('('));
					}
					if (aDialect.Functions.Contains(lFunctionName)) 
					{
						// this is function, don't quote it
						return lDefaultValue;
					}
					lDefaultValue = lDefaultValue.Replace("'", "''");
					lDefaultValue = "'" + lDefaultValue + "'";
				}
			} 

			return lDefaultValue;
		}

		public string[] generateComment(EREntity aEntity, Dialect aDialect) 
		{
			ArrayList lSQL = new ArrayList();
			if (aDialect.SupportsCommentOn) 
			{
				int lQuoteSQLIdentifier = aEntity.GetModel().getQuoteSQLIdentifier();
				string tableName = DDLGeneratorUtil.GetQualifiedName(aEntity.Schema, aEntity.GetName(), aDialect, lQuoteSQLIdentifier);
				string comment = aEntity.Comment;
				if (comment != null && comment.Length > 0) 
				{
					comment = comment.Replace("'", "''");
					StringBuilder buf = new StringBuilder().Append("comment on table ").Append(tableName)
						.Append(" is '").Append(comment).Append("'");
					lSQL.Add(buf.ToString());
				}
				foreach (ERElement element in aEntity.Elements)
				{
					if (element is ERColumn) 
					{
						ERColumn column = (ERColumn) element;
						String columnComment = column.Comment;
						if (columnComment != null && columnComment.Length > 0) 
						{
							StringBuilder buf = new StringBuilder().Append("comment on column ").Append(
								tableName).Append('.').Append(DDLGeneratorUtil.QuoteName(column.GetName(), aDialect, lQuoteSQLIdentifier)).Append(
								" is '").Append(columnComment).Append("'");
							lSQL.Add(buf.ToString());
						}
					}
				}
			}
			return lSQL.ToArray(typeof(string)) as string[];
		}

	}
}
