using System;
using Orm.Util;
using Orm.Cfg;
using NHibernate;
using NHibernate.Cfg;
using NHibernate.Dialect;
using NHibernate.Mapping;
using NHibernate.Id;
using System.Collections;
using System.Runtime.CompilerServices;

namespace Orm
{
	public class PersistentManager
	{
		public virtual string GetProjectName()
		{	
			return null;
		}

		public virtual System.Reflection.Assembly GetAssembly()
		{
			return null;
		}

		public enum SessionType 
		{
			APP_BASE,
			THREAD_BASE,
			USER_DEFINIED
		}
		
		
		private static SessionType _sessionType = SessionType.APP_BASE;
		private SessionController _sessionController;
		private Configuration _cfg;
		private FlushMode _flushMode = FlushMode.Unspecified;

		public PersistentManager(DotNetConnectionSetting connectionSetting, SessionType sessionType)
		{
			_sessionType = sessionType;
			_cfg = new Configuration();
			try 
			{
				_cfg.Configure();
			} 
			catch (Exception) 
			{
				// Try to load from assembly
				try 
				{
					_cfg.Configure(GetAssembly(), "hibernate.cfg.xml");
				}
				catch (Exception) 
				{
					System.Reflection.Assembly assembly = GetAssembly();
					string[] resources = assembly.GetManifestResourceNames();
					foreach (string resource in resources) 
					{
						if (resource.EndsWith("hibernate.cfg.xml")) 
						{
							_cfg.Configure(assembly, resource);
							break;
						}
					}
				}
			}

			if(connectionSetting != null)
			{
				_cfg.SetProperty("hibernate.connection.provider", connectionSetting.ConnectionProvider);
				_cfg.SetProperty("hibernate.dialect", connectionSetting.Dialect);
				_cfg.SetProperty("hibernate.connection.driver_class", connectionSetting.DriverClass);
				_cfg.SetProperty("hibernate.connection.connection_string", connectionSetting.ConnectionString);
			}
			ResolveNativeIDGenerators();

			if(_sessionType == SessionType.APP_BASE)
			{
				_sessionController = new AppBaseSessionController(_cfg);
			}
			else if(_sessionType == SessionType.THREAD_BASE)
			{
				_sessionController = new ThreadBaseSessionController(_cfg);
			}
			else if(_sessionType == SessionType.USER_DEFINIED)
			{
				_sessionController = new UserDefinedSessionController(_cfg);
			}
		}

		private void ResolveNativeIDGenerators()
		{
			Dialect dialect = Dialect.GetDialect(_cfg.Properties);
			ICollection classMappings = _cfg.ClassMappings;
			foreach(PersistentClass pc in classMappings)
			{
				if(!pc.IsInherited)
				{
					if(pc.Identifier.GetType().Equals(typeof(SimpleValue)))
					{
						SimpleValue identifier = pc.Identifier as SimpleValue;
						ResolveIdentifier(identifier, dialect);
					}
				}
			}

			ICollection collectionMappings = _cfg.CollectionMappings;
			foreach(Collection collection in collectionMappings)
			{
				if(collection.IsIdentified)
				{
					if(((IdentifierCollection)collection).Identifier.GetType().Equals((typeof(SimpleValue))))
					{
						SimpleValue identifier = (SimpleValue)((IdentifierCollection)collection).Identifier;
						ResolveIdentifier(identifier, dialect);
					}
				}
			}
		}


		private void ResolveIdentifier(SimpleValue identifier, Dialect dialet)
		{
			if(identifier.IdentifierGeneratorStrategy.Equals("native"))
			{
				String generator = DDLGeneratorUtil.GETIDGererator("native", dialet, isIntLongShort(identifier.Type.Name)); 
				identifier.IdentifierGeneratorStrategy = generator;
				if(generator.Equals("assigned"))
				{
					identifier.NullValue = "undefined";
				}
				else
				{
					identifier.NullValue = null;
					Column lColumn = null;
					foreach(Column idColumn in identifier.ColumnCollection)
					{
						lColumn = idColumn;
					}
					string idGeneratorKey = DDLGeneratorUtil.GetIDGeneratorKey(identifier.Table.Schema, identifier.Table.Name, lColumn.Name, generator, null, dialet,
						identifier.Table.IsQuoted, lColumn.IsQuoted);
					if(generator.Equals("sequence"))
					{
						identifier.IdentifierGeneratorProperties.Add(SequenceGenerator.Sequence, idGeneratorKey);
					}
					else if(generator.Equals("hilo"))
					{
						identifier.IdentifierGeneratorProperties.Add(TableGenerator.Table, idGeneratorKey);
					}
					else
					{
						identifier.IdentifierGeneratorProperties.Add(TableGenerator.Column, idGeneratorKey); 
					}
						
				}
			}
		}
		

		private bool isIntLongShort(string typeName)
		{
			if(typeName.Equals("Int32") || typeName.Equals("Int16") || typeName.Equals("Int64"))
			{
				return true;
			}
			return false;
		}
//		public static void SetSessionType(SessionType sessionType)  
//		{
//			if (_instance != null) 
//			{
//				throw new PersistentException("Cannot set session type after create PersistentManager instance");
//			} 
//			else 
//			{
//				_sessionType = sessionType;
//			}
//		}
		
		public PersistentSession GetSession()
		{
			try
			{
				PersistentSession session = _sessionController.GetSession();
				if (session!=null) {
					session.FlushMode = _flushMode;
				}
				return session;
			}
			catch(HibernateException e)
			{
				throw new PersistentException(e);
			}
		}

		public void DisposePersistentManager()
		{
			_sessionController.DisposeSession();
		}

		public void SaveObject(object obj)
		{
			PersistentSession session = GetSession();
			if(!session.IsTransaction || _sessionType == SessionType.USER_DEFINIED)
			{
				PersistentTransaction transaction = session.BeginTransaction();
				session.SaveOrUpdate(obj);
				transaction.Commit();
			}
			else 
			{
				session.SaveOrUpdate(obj);
			}

			if(_sessionType == SessionType.USER_DEFINIED)
			{
				session.Close();
			}
		}

		public void DeleteObject(object obj)
		{
			PersistentSession session = GetSession();
			if(!session.IsTransaction || _sessionType == SessionType.USER_DEFINIED)
			{
				PersistentTransaction transaction = session.BeginTransaction();
				session.Delete(obj);
				transaction.Commit();
			}
			else
			{
				session.Delete(obj);
			}

			if(_sessionType == SessionType.USER_DEFINIED)
			{
				session.Close();
			}
		}

//		[MethodImpl(MethodImplOptions.Synchronized)]
//		public static PersistentManager Instance()  
//		{
//			if (_instance == null) 
//			{
//				_instance = new PersistentManager();
//			}
//
//			return _instance;
//		}

		public Configuration GetCfg
		{
			get { return _cfg; }
		}
		
		public FlushMode FlushMode {
			get {
				return _flushMode;
			}
			
			set {
				_flushMode = value;
			}
		}

        public IInterceptor Interceptor
        {
            get
            {
                return _sessionController.Interceptor.UserInterceptor;
            }

            set
            {
                _sessionController.Interceptor.UserInterceptor = value;
            }
        }
	}
}
