using System;
using System.Reflection;
using NHibernate;
using NHibernate.Cfg;
using NHibernate.Mapping;
using NHibernate.Type;

namespace Orm.Util
{
	/// <summary>
	/// Summary description for ORMInterceptor.
	/// </summary>
	public class ORMInterceptor : EmptyInterceptor
	{
		private readonly Configuration _configuration;
		private static readonly log4net.ILog log = log4net.LogManager.GetLogger(typeof(ORMInterceptor));
        private IInterceptor _userInterceptor;

        public IInterceptor UserInterceptor
        {
            get
            {
                return _userInterceptor;
            }

            set
            {
                _userInterceptor = value;
            }
        }

		public ORMInterceptor(Configuration aConfiguration) 
		{
			_configuration = aConfiguration;
		}

		#region IInterceptor Members

        public override bool OnLoad(object entity, object id, object[] state, string[] propertyNames, IType[] types)
		{
			PersistentClass lClassMapping = _configuration.GetClassMapping(entity.GetType());
			if (lClassMapping.Identifier.IdentifierGeneratorStrategy.Equals("assigned")) 
			{
				try 
				{
					MethodInfo lMethod = lClassMapping.MappedClass.GetMethod("onLoad");
					lMethod.Invoke(entity, null);                
				} 
				catch (Exception e) 
				{
					log.Warn("Exception on Interceptor.OnLoad for class: " + entity.GetType().FullName, e);
				}
			}
            if (_userInterceptor != null)
            {
                return _userInterceptor.OnLoad(entity, id, state, propertyNames, types);
            }
            else
            {
                return false;
            }
		}

        public override bool OnSave(object entity, object id, object[] state, string[] propertyNames, IType[] types)
		{
			PersistentClass lClassMapping = _configuration.GetClassMapping(entity.GetType());
			if (lClassMapping.Identifier.IdentifierGeneratorStrategy.Equals("assigned")) 
			{
				try 
				{
					MethodInfo lMethod = lClassMapping.MappedClass.GetMethod("onSave");
					lMethod.Invoke(entity, null);                
				} 
				catch (Exception e) 
				{
					log.Warn("Exception on Interceptor.OnSave for class: " + entity.GetType().FullName, e);
				}
			}
            if (_userInterceptor != null)
            {
                return _userInterceptor.OnSave(entity, id, state, propertyNames, types);
            }
            else
            {
                return false;
            }
		}

        public override object IsUnsaved(object entity)
		{
			PersistentClass lClassMapping = _configuration.GetClassMapping(entity.GetType());
			if (lClassMapping.Identifier.IdentifierGeneratorStrategy.Equals("assigned")) 
			{
				try 
				{
					MethodInfo lMethod = lClassMapping.MappedClass.GetMethod("isSaved");
					bool lSaved = (bool) lMethod.Invoke(entity, null);
					return !lSaved;
				} 
				catch (Exception e) 
				{
					log.Warn("Exception on Interceptor.IsUnsaved for class: " + entity.GetType().FullName, e);
				}
			}
            if (_userInterceptor != null)
            {
                return _userInterceptor.IsUnsaved(entity);
            }
            else
            {
                return null;
            }
		}

        public override void AfterTransactionBegin(ITransaction tx)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.AfterTransactionBegin(tx);
            }
            base.AfterTransactionBegin(tx);
        }

        public override void AfterTransactionCompletion(ITransaction tx)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.AfterTransactionCompletion(tx);
            }
            base.AfterTransactionCompletion(tx);
        }

        public override void BeforeTransactionCompletion(ITransaction tx)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.BeforeTransactionCompletion(tx);
            }
            base.BeforeTransactionCompletion(tx);
        }

        public override int[] FindDirty(object entity, object id, object[] currentState, object[] previousState, string[] propertyNames, IType[] types)
        {
            if (_userInterceptor != null)
            {
                return _userInterceptor.FindDirty(entity, id, currentState, previousState, propertyNames, types);
            }
            return base.FindDirty(entity, id, currentState, previousState, propertyNames, types);
        }

        public override object Instantiate(Type clazz, object id)
        {
            if (_userInterceptor != null)
            {
                return _userInterceptor.Instantiate(clazz, id);
            }
            return base.Instantiate(clazz, id);
        }

        public override void OnDelete(object entity, object id, object[] state, string[] propertyNames, IType[] types)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.OnDelete(entity, id, state, propertyNames, types);
            }
            base.OnDelete(entity, id, state, propertyNames, types);
        }

        public override bool OnFlushDirty(object entity, object id, object[] currentState, object[] previousState, string[] propertyNames, IType[] types)
        {
            if (_userInterceptor != null)
            {
                return _userInterceptor.OnFlushDirty(entity, id, currentState, previousState, propertyNames, types);
            }
            return base.OnFlushDirty(entity, id, currentState, previousState, propertyNames, types);
        }

        public override void PostFlush(System.Collections.ICollection entities)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.PostFlush(entities);
            }
            base.PostFlush(entities);
        }

        public override void PreFlush(System.Collections.ICollection entitites)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.PreFlush(entitites);
            }
            base.PreFlush(entitites);
        }

        public override void SetSession(ISession session)
        {
            if (_userInterceptor != null)
            {
                _userInterceptor.SetSession(session);
            }
            base.SetSession(session);
        }

		#endregion
	}
}
